package org.xiss.index.impl;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sf.xanswer.indexing.IndexException;
import net.sf.xanswer.storage.map.PersistentException;

import org.xiss.index.ElementIndexEntry;
import org.xiss.index.IXissIndex;
import org.xiss.index.StructureIndexEntry;
import org.xiss.index.impl.persistence.PersistentList;
import org.xiss.map.BTreeFactory;
import org.xiss.map.IBTree;
import org.xiss.map.PersistentBTreeFactory;
import org.xiss.query.IQuery;
import org.xiss.query.IResult;

class XissIndexImpl implements IXissIndex {
	private static final String VALUE = "VALUE";

	private static final String VID = "VID";

	private static final String NAME = "NAME";

	private static final String NID = "NID";

	private static final String STRUCTURE = "STRUCTURE";

	private static final String AINDEX = "AINDEX";

	private static final String EINDEX = "EINDEX";

	private static final int EINDEX_CACHE_CAPACITY = 16;

	private static final int AINDEX_CACHE_CAPACITY = 16;

	private static final int NID_CACHE_CAPACITY = 1024;

	private static final int VID_CACHE_CAPACITY = 1024;

	private static final int NAME_CACHE_CAPACITY = 1024;

	private static final int VALUE_CACHE_CAPACITY = 1024;

	// Key: NID
	private IBTree<Integer, List<ElementIndexEntry>> myElementIndex;

	// Key: NID
	private IBTree<Integer, List<ElementIndexEntry>> myAttributeIndex;

	private PersistentList<StructureIndexEntry> myStructureIndex;

	// Key: Name Value: NID
	private IBTree<String, Integer> myNidTable;

	// Key: NID Value: Name
	private IBTree<Integer, String> myNameTable;

	private int myDistinctNames;

	// Key: value Value: VID
	private IBTree<String, Integer> myVidTable;

	// Key: VID Value: value
	private IBTree<Integer, String> myValueTable;

	private int myDistinctValues;

	private int myTotalCount;

	private ElementIndexEntry myRootElementIndexEntry;

	private XissIndexImpl() {
	}

	static XissIndexImpl openExisting(String inputDirectory)
			throws IndexException {
		XissIndexImpl impl = new XissIndexImpl();
		impl.constructFromExisting(inputDirectory);
		return impl;
	}

	static XissIndexImpl createNew(String outputDirectory)
			throws IndexException {
		XissIndexImpl impl = new XissIndexImpl();
		impl.constructNew(outputDirectory);
		return impl;
	}

	public int updateNameTable(String name) throws IndexException {
		try {
			Integer nid = myNidTable.get(name);
			if (nid != null) {
				return nid;
			}
			myNidTable.put(name, myDistinctNames);
			myNameTable.put(myDistinctNames, name);
			return myDistinctNames++;
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public int updateValueTable(String value) throws IndexException {
		try {
			Integer vid = myVidTable.get(value);
			if (vid != null) {
				return vid;
			}
			myVidTable.put(value, myDistinctValues);
			myValueTable.put(myDistinctValues, value);

			return myDistinctValues++;
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public IBTree<Integer, List<ElementIndexEntry>> getElementIndex()
			throws IndexException {
		return myElementIndex;
	}

	public List<ElementIndexEntry> getElementIndexEntriesByNid(int nid)
			throws IndexException {
		try {
			if(nid == NONE)
				return new ArrayList<ElementIndexEntry>(0);
			return Collections.unmodifiableList(myElementIndex.get(nid));
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public IBTree<Integer, List<ElementIndexEntry>> getAttributeIndex()
			throws IndexException {
		return myAttributeIndex;
	}

	public List<ElementIndexEntry> getAttributeIndexEntriesByNid(int nid)
			throws IndexException {
		try {
			if(nid == NONE)
				return new ArrayList<ElementIndexEntry>(0);
			return Collections.unmodifiableList(myAttributeIndex.get(nid));
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public List<StructureIndexEntry> getStructureIndex() throws IndexException {
		return Collections.unmodifiableList(myStructureIndex);
	}

	public ElementIndexEntry getRootElementIndexEntry() throws IndexException {
		return myRootElementIndexEntry;
	}

	void addToAttributeIndex(int nid, ElementIndexEntryImpl entry)
			throws IndexException {
		try {
			List<ElementIndexEntry> sameNameAttributes = myAttributeIndex
					.get(nid);
			if (sameNameAttributes == null) {
				sameNameAttributes = new ArrayList<ElementIndexEntry>(1);
			}

			sameNameAttributes.add(entry);
			entry.setEnclosingList(sameNameAttributes);
			myAttributeIndex.put(nid, sameNameAttributes);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	void addElementIndexEntry(int nid, ElementIndexEntryImpl elementEntry,
			boolean isRoot) throws IndexException {
		try {
			if (isRoot) {
				myRootElementIndexEntry = elementEntry;
			}
			List<ElementIndexEntry> list = myElementIndex.get(nid);
			if (list == null) {
				list = new ArrayList<ElementIndexEntry>(1);
			}
			list.add(elementEntry);
			elementEntry.setEnclosingList(list);
			elementEntry.setIndex(list.size() - 1);
			myElementIndex.put(nid, list);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	void updateElementIndexEntry(int nid, ElementIndexEntryImpl elementEntry)
			throws IndexException {
		try {
			List<ElementIndexEntry> list = myElementIndex.get(nid);
			list.set(elementEntry.getIndex(), elementEntry);
			myElementIndex.put(nid, list);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	void addStructureIndexEntry(StructureIndexEntryImpl structureEntry)
			throws IndexException {
		myStructureIndex.add(structureEntry);
	}

	public int getTotalNodesCount() throws IndexException {
		return myTotalCount;
	}

	void incrementNodeCount() {
		++myTotalCount;
	}

	public String getNodeName(int id) throws IndexException {
		try {
			return myNameTable.get(myStructureIndex.get(id).NameId);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public String getNodeValue(int id) throws IndexException {
		try {
			return myNameTable.get(myStructureIndex.get(id).ValueId);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public IBTree<Integer, String> getNameTable() {
		return myNameTable;
	}

	public int getNidByName(String name) throws IndexException {
		try {
			Integer res = myNidTable.get(name);
			if(res == null)
				return NONE;
			return myNidTable.get(name);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public String getNameByNid(int nid) throws IndexException {
		try {
			if(nid == NONE)
				return null;
			return myNameTable.get(nid);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public IBTree<Integer, String> getValueTable() throws IndexException {
		return myValueTable;
	}

	public int getVidByValue(String value) throws IndexException {
		try {
			Integer res = myNidTable.get(value);
			if(res == null)
				return NONE;
			return myVidTable.get(value);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public String getValueByVid(int vid) throws IndexException {
		try {
			if(vid == NONE)
				return null;
			return myValueTable.get(vid);
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public IResult query(IQuery query) throws IndexException {
		return query.run(this);
	}

	private void constructFromExisting(String inputDirectory)
			throws IndexException {

		try {
			File dir = new File(inputDirectory);
			if (!dir.exists() || !dir.isDirectory()) {
				throw new IndexException("Directory " + inputDirectory
						+ " does not exist");
			}

			BTreeFactory<Integer, List<ElementIndexEntry>> factory1 = new PersistentBTreeFactory<Integer, List<ElementIndexEntry>>();
			BTreeFactory<String, Integer> factory2 = new PersistentBTreeFactory<String, Integer>();
			BTreeFactory<Integer, String> factory3 = new PersistentBTreeFactory<Integer, String>();
			myElementIndex = factory1.open(inputDirectory,
					XissIndexImpl.EINDEX, EINDEX_CACHE_CAPACITY);
			myAttributeIndex = factory1.open(inputDirectory,
					XissIndexImpl.AINDEX, AINDEX_CACHE_CAPACITY);
			// TODO:
			// Structure index persistence is now implemented with brute force
			myStructureIndex = new PersistentList<StructureIndexEntry>();
			myStructureIndex.create(inputDirectory, XissIndexImpl.STRUCTURE);
			myStructureIndex.open();

			myNidTable = factory2.open(inputDirectory, XissIndexImpl.NID,
					NID_CACHE_CAPACITY);
			myNameTable = factory3.open(inputDirectory, XissIndexImpl.NAME,
					NAME_CACHE_CAPACITY);
			myVidTable = factory2.open(inputDirectory, XissIndexImpl.VID,
					VID_CACHE_CAPACITY);
			myValueTable = factory3.open(inputDirectory, XissIndexImpl.VALUE,
					VALUE_CACHE_CAPACITY);

			myElementIndex.open();
			myAttributeIndex.open();
			myNidTable.open();
			myNameTable.open();
			myVidTable.open();
			myValueTable.open();
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	private void constructNew(String directory) throws IndexException {
		try {
			File dir = new File(directory);
			if (!dir.exists()) {
				dir.mkdir();
			} else if (!dir.isDirectory()) {
				throw new IndexException(directory + " is not a directory");
			}

			BTreeFactory<Integer, List<ElementIndexEntry>> factory1 = new PersistentBTreeFactory<Integer, List<ElementIndexEntry>>();
			BTreeFactory<String, Integer> factory2 = new PersistentBTreeFactory<String, Integer>();
			BTreeFactory<Integer, String> factory3 = new PersistentBTreeFactory<Integer, String>();
			myElementIndex = factory1.create(directory, XissIndexImpl.EINDEX,
					EINDEX_CACHE_CAPACITY);
			myAttributeIndex = factory1.create(directory, XissIndexImpl.AINDEX,
					AINDEX_CACHE_CAPACITY);
			myStructureIndex = new PersistentList<StructureIndexEntry>();
			myStructureIndex.create(directory, XissIndexImpl.STRUCTURE);
			try {
				myStructureIndex.createNew();
			} catch (PersistentException e) {
				throw new IndexException(e);
			}

			myNidTable = factory2.create(directory, XissIndexImpl.NID,
					NID_CACHE_CAPACITY);
			myNameTable = factory3.create(directory, XissIndexImpl.NAME,
					NAME_CACHE_CAPACITY);
			myVidTable = factory2.create(directory, XissIndexImpl.VID,
					VID_CACHE_CAPACITY);
			myValueTable = factory3.create(directory, XissIndexImpl.VALUE,
					VALUE_CACHE_CAPACITY);

			myTotalCount = 0;
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public void close() throws IndexException {
		try {
			myElementIndex.close();
			myAttributeIndex.close();
			myNidTable.close();
			myNameTable.close();
			myVidTable.close();
			myValueTable.close();
			myStructureIndex.close();
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}

	public void drop() throws IndexException {
		try {
			myElementIndex.drop();
			myAttributeIndex.drop();
			myNidTable.drop();
			myNameTable.drop();
			myVidTable.drop();
			myValueTable.drop();
			myStructureIndex.drop();
		} catch (PersistentException e) {
			throw new IndexException(e);
		}
	}
}
