package org.xiss.query.impl;

import java.util.HashMap;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class AdvancedTokenizer {
	private String myString;

	private Stack<Integer> myPositionStack;

	private boolean myIgnoreWhitespace;

	private HashMap<String, Pattern> myPatternCache;

	public AdvancedTokenizer(String string) {
		myString = string;
		myPositionStack = new Stack<Integer>();
		myPositionStack.push(0);
		myPatternCache = new HashMap<String, Pattern>();
		myIgnoreWhitespace = false;
	}

	/**
	 * Sets or clears whitespace ignoring mode. In whitespace ignoring mode, all
	 * functions except for {@link #eatWhitespace()} and
	 * {@link #eatWhitespace(boolean)} first eat whitespace and only after do
	 * their job.
	 * 
	 */
	public void setIgnoreWhitespace(boolean ignoreWhitespace) {
		myIgnoreWhitespace = true;
	}

	/**
	 * Returns whether the tokenizer is ignoring whitespace.
	 * 
	 * @see #setIgnoreWhitespace(boolean)
	 */
	public boolean isIgnoringWhitespace() {
		return myIgnoreWhitespace;
	}

	private void maybeEatWhitespace() {
		if (myIgnoreWhitespace) {
			eatWhitespace();
		}
	}

	/**
	 * Eats a (possibly empty) whitespace sequence starting from the current
	 * position.
	 * 
	 */
	public void eatWhitespace() {
		eatWhitespace(false);
	}

	/**
	 * Eats a whitespace sequence starting from the current position
	 * 
	 * @param isRequired
	 *            If true, then at least one whitespace is required (in case no
	 *            whitespace is found, an exception is thrown).
	 */
	public void eatWhitespace(boolean isRequired) {
		boolean whitespacePresent = false;
		while (getPosition() < getLength()
				&& Character.isWhitespace(myString.charAt(getPosition()))) {
			whitespacePresent = true;
			incPosition();
		}
		if (isRequired && !whitespacePresent) {
			throw new LexerException(getPosition(), "Whitespace required");
		}
	}

	/**
	 * Eats a token starting from the current position. Post-condition: Position
	 * points to the first symbol after the expected token, or is left unchanged
	 * if the token was not found
	 * 
	 * @param token
	 *            The expected token
	 * @return true if the token was found, false otherwise
	 */
	public boolean eat(String token, boolean isRequired) {
		maybeEatWhitespace();
		int i;
		if (myString.length() - getPosition() < token.length()) {
			if (isRequired) {
				throw new LexerException(getPosition(), "'" + token
						+ "' expected");
			}
			return false;
		}
		for (i = 0; i < token.length(); ++i) {
			if (myString.charAt(i + getPosition()) != token.charAt(i)) {
				if (isRequired) {
					throw new LexerException(getPosition(), "'" + token
							+ "' expected");
				}
				return false;
			}
		}
		incPosition(token.length());
		return true;
	}

	/**
	 * Eats a character just like a regular token
	 * 
	 * @see #eat(String)
	 */
	public boolean eat(char c) {
		maybeEatWhitespace();
		if (myString.length() == getPosition()) {
			return false;
		}
		if (myString.charAt(getPosition()) != c) {
			return false;
		}
		incPosition();
		return true;
	}

	/**
	 * Eats a character just like {@link #eat(char)} and possibly throws an
	 * exception if it is not found
	 * 
	 * @param c
	 *            The seeked character
	 * @param isRequired
	 *            If true, an exception is thrown in case the character is not
	 *            found
	 */
	public boolean eat(char c, boolean isRequired) {
		boolean res = eat(c);
		if (!res && isRequired) {
			throw new LexerException(getPosition(), "'" + c + "' expected");
		}
		return res;
	}

	/**
	 * Eats a character from the given subset
	 * 
	 * @see #eat(char)
	 */
	public boolean eatAny(String chars) {
		maybeEatWhitespace();
		if (myString.length() == getPosition()) {
			return false;
		}
		if (chars.indexOf(myString.charAt(getPosition())) < 0) {
			return false;
		}
		incPosition();
		return true;
	}

	/**
	 * Eats a character from the given subset and possibly throws an exception
	 * if it is not found
	 * 
	 * @param c
	 *            The seeked set of characters
	 * @param isRequired
	 *            If true, an exception is thrown in case a fitting character is
	 *            not found
	 */
	public boolean eatAny(String chars, boolean isRequired) {
		boolean res = eatAny(chars);
		if (!res && isRequired) {
			throw new LexerException(getPosition(), "Any of: '" + chars
					+ "' expected");
		}
		return res;
	}

	/**
	 * Reads the string until whitespace is encountered. Post-condition:
	 * Position points to the first whitespace character after the previous
	 * position value
	 * 
	 * @return Portion of the string starting at the current position and ending
	 *         just before the next whitespace symbol
	 */
	public String readToWhitespace() {
		maybeEatWhitespace();
		String rem = getRemaining();
		int idx = AdvancedTokenizer.getFirstWhitespacePosition(rem);
		incPosition(idx);
		return (idx < 0) ? rem : (rem.substring(0, idx));
	}

	/**
	 * Reads the string, maybe skipping starting whitespace, until the next
	 * whitespace is encountered
	 * 
	 * @param canStartWithWhitespace
	 *            If true, then {@link #eatWhitespace()} is called before
	 *            reading the result, otherwise it is called only if
	 *            ignoringWhitespace ({@link #setIgnoreWhitespace(boolean)})
	 *            is set to true.
	 * @see #readToWhitespace()
	 */
	public String readToWhitespace(boolean canStartWithWhitespace) {
		maybeEatWhitespace();
		if (canStartWithWhitespace) {
			eatWhitespace();
		}
		return readToWhitespace();
	}

	/**
	 * Reads portion of the string starting at the current position until the
	 * given character is encountered, or to the end of the string if it is not
	 * found Post-condition: Position points to the first occurence of the given
	 * character if it is found, otherwise position is left unchanged
	 * 
	 * @param c
	 *            The expected character
	 * @param isRequired
	 *            If true, an exception is thrown in case the character is not
	 *            found
	 */
	public String readUntil(char c, boolean isRequired) {
		maybeEatWhitespace();
		String rem = getRemaining();
		int idx = rem.indexOf(c);
		if (idx >= 0) {
			incPosition(idx);
			return rem.substring(0, idx);
		}
		if (!isRequired) {
			return rem;
		}
		throw new LexerException(getLength(), "'" + c + "' expected");
	}

	/**
	 * Same as {@link #readUntil(char, boolean)} but seeks for any character
	 * from the given set, not a single one. Post-condition: Position points to
	 * the first character starting from the old position belonging to the given
	 * set, or to the end of the string if it is not found.
	 * 
	 * @param chars
	 *            The seeked character set
	 * @param isRequired
	 *            If true, then an exception is thrown in case the remaining
	 *            portion of the string doesn't contain a character from the
	 *            given set.
	 * @see #readUntil(char, boolean)
	 */
	public String readUntilAny(String chars, boolean isRequired) {
		maybeEatWhitespace();
		int pos = getPosition();
		for (; pos < myString.length(); ++pos) {
			if (chars.indexOf(myString.charAt(pos)) >= 0) {
				break;
			}
		}
		if (pos != myString.length()) {
			String res = myString.substring(getPosition(), pos);
			setPosition(pos);
			return res;
		}
		if (isRequired) {
			throw new LexerException(getLength(), "One of: '" + chars
					+ "' expected");
		}
		return getRemaining();
	}

	/**
	 * Same as {@link #readUntilAny(String, boolean)} but also stops if a
	 * whitespace is encountered
	 */
	public String readUntilAnyOrWhitespace(String chars, boolean isRequired) {
		maybeEatWhitespace();
		int pos = getPosition();
		for (; pos < myString.length(); ++pos) {
			if (Character.isWhitespace(myString.charAt(pos))
					|| chars.indexOf(myString.charAt(pos)) >= 0) {
				break;
			}
		}
		if (pos != myString.length()) {
			String res = myString.substring(getPosition(), pos);
			setPosition(pos);
			return res;
		}
		if (isRequired) {
			throw new LexerException(getLength(), "One of: '" + chars
					+ "' expected");
		}
		return getRemaining();

	}

	/**
	 * Reads the longest portion of the remaining string consisting of
	 * characters from the given set. Post-condition: Position points to the
	 * first character not in the given set, or to the end of the string if such
	 * a character is not found (all remaining characters belong to the set)
	 * 
	 * @param chars
	 *            Character set
	 */
	public String readWhileAny(String chars) {
		maybeEatWhitespace();
		int pos = getPosition();
		for (; pos < myString.length(); ++pos) {
			if (chars.indexOf(myString.charAt(pos)) < 0) {
				break;
			}
		}
		String res = myString.substring(getPosition(), pos);
		setPosition(pos);
		return res;
	}

	/**
	 * Reads portion of the string from the current position to position of the
	 * given token, or to the end if the token is not found. Post-condition:
	 * Position points to the first occurence of the given token, or to the end
	 * of the string if the token is not found.
	 * 
	 * @param token
	 *            The seeked token
	 * @param isRequired
	 *            If true, an exception is thrown in case the token is not found
	 */
	public String readUntilToken(String token, boolean isRequired) {
		maybeEatWhitespace();
		String rem = getRemaining();
		int idx = rem.indexOf(token);
		if (idx >= 0) {
			incPosition(idx);
			return rem.substring(0, idx);
		}
		if (!isRequired) {
			return rem;
		}
		throw new LexerException(getLength(), "'" + token + "' expected");
	}

	/**
	 * Reads portion of the string from the current position to position of the
	 * first match to a regex, or to the end if no match is found.
	 * Post-condition: Position points to the first match, or to the end of the
	 * string if no match is found Note: Each pattern is compiled only once and
	 * patterns are stored in a pattern cache that is never cleared.
	 * 
	 * @param pattern
	 *            The seeked pattern
	 * @param isRequired
	 *            If true, an exception is thrown in case the token is not found
	 */
	public String readUntilRegex(String pattern, boolean isRequired) {
		maybeEatWhitespace();
		Pattern p = getPattern(pattern);
		String rem = getRemaining();
		Matcher m = p.matcher(rem);
		if (!m.find()) {
			if (isRequired) {
				throw new LexerException(getLength(), "Regex-match '" + pattern
						+ "' expected");
			}
			return rem;
		}
		String res = rem.substring(0, m.start());
		incPosition(m.start());
		return res;
	}

	/**
	 * Eats portion of the string from the current position that matches the
	 * given regex. Post-condition: Position points to the portion of the string
	 * remaining after cutting off the matched prefix.
	 * 
	 * @param pattern
	 *            The seeked pattern
	 * @param isRequired
	 *            If true, an exception is thrown in case a match is not found.
	 * @return The matched prefix, or null if a match is not found and
	 *         isRequired == false.
	 */
	public String eatRegex(String pattern, boolean isRequired) {
		maybeEatWhitespace();
		Pattern p = getPattern(pattern);
		String rem = getRemaining();
		Matcher m = p.matcher(rem);
		if (!m.find() || m.start() != 0) {
			if (isRequired) {
				throw new LexerException(getPosition(), "Regex-match '"
						+ pattern + "' expected");
			} else {
				return null;
			}
		}
		String res = rem.substring(0, m.end());
		incPosition(m.end());
		return res;
	}

	/**
	 * Returns portion of the string from the current position to the end
	 */
	public String getRemaining() {
		return myString.substring(getPosition());
	}

	/**
	 * Begins a match transaction: the current position is stored into a stack.
	 * Example:
	 * 
	 * <pre>
	 *     	boolean matchSomeExtremelyComlexSequence(AdvancedTokenizer tokenizer) {
	 *     		tokenizer.begin();
	 *     		if(matchFirstVeryComplexAlternative(tokenizer)) {
	 *     			tokenizer.commit();
	 *     		} else if(tokenizer.retry() &amp;&amp; matchSecondVeryComplexAlternative(tokenizer)) {
	 *    	 		tokenizer.commit();
	 *     		} ... else {
	 *     			tokenizer.rollback();
	 *     			return false;	
	 *     		}
	 *     		return true;
	 *     	}
	 * </pre>
	 * 
	 * @see #commit()
	 * @see #rollback()
	 */
	public void begin() {
		myPositionStack.push(getPosition());
	}

	/**
	 * Retries a match transaction. Does {@link #rollback()} and then
	 * {@link #begin()}.
	 * 
	 * @return Always true (Syntactic sugar: see {@link #begin()})
	 * @see #begin()
	 */
	public boolean retry() {
		rollback();
		begin();
		return true;
	}

	/**
	 * Commits a match transaction
	 * 
	 */
	public void commit() {
		Integer val = myPositionStack.pop();
		setPosition(val);
	}

	/**
	 * Cancels a match transaction
	 * 
	 */
	public void rollback() {
		myPositionStack.pop();
	}

	/**
	 * Returns whether end of the string has been reached
	 */
	public boolean isDone() {
		return getPosition() == myString.length();
	}

	private int getLength() {
		return myString.length();
	}

	private int getPosition() {
		return myPositionStack.peek();
	}

	private void setPosition(int pos) {
		myPositionStack.set(myPositionStack.size() - 1, pos);
	}

	private void incPosition() {
		setPosition(getPosition() + 1);
	}

	private void incPosition(int step) {
		setPosition(getPosition() + step);
	}

	private static int getFirstWhitespacePosition(String str) {
		for (int i = 0; i < str.length(); ++i) {
			if (Character.isWhitespace(str.charAt(i))) {
				return i;
			}
		}
		return -1;
	}

	private Pattern getPattern(String pattern) {
		Pattern p;
		if (myPatternCache.containsKey(pattern)) {
			p = myPatternCache.get(pattern);
		} else {
			p = Pattern.compile(pattern);
		}
		return p;
	}

	private static class LexerException extends RuntimeException {
		private static final long serialVersionUID = 1L;

		public LexerException(int pos, String message) {
			super("Illegal token at position " + pos + ": " + message);
		}
	}
}
