package org.xiss.query.impl;

import java.util.ArrayList;
import java.util.List;

import org.xiss.query.IXissOperation;
import org.xiss.query.QueryParseException;
import org.xiss.query.impl.ast.AnyNode;
import org.xiss.query.impl.ast.AstNode;
import org.xiss.query.impl.ast.AttributeChunkNode;
import org.xiss.query.impl.ast.AttributePredicateNode;
import org.xiss.query.impl.ast.ElementChunkNode;
import org.xiss.query.impl.ast.ElementPredicateNode;
import org.xiss.query.impl.ast.GlobalPathNode;
import org.xiss.query.impl.ast.PathChunkNode;
import org.xiss.query.impl.ast.PathNode;
import org.xiss.query.impl.ast.PredicateNode;
import org.xiss.query.impl.ast.UnionNode;

public class XissQueryParser {
	private AdvancedTokenizer myTokenizer;

	private List<IXissOperation> myScript;

	private String myQuery;

	private static String NAME_REGEX = "(text\\(\\))|([^_][a-zA-Z:0-9_]*)";

	private static String ATTRIBUTE_VALUE_REGEX = "[^\"']*";

	public XissQueryParser(String query) {
		myQuery = query;
	}

	public XissQuery parse() throws QueryParseException {
		myTokenizer = new AdvancedTokenizer(myQuery);

		AstNode node = parseQuery(true);
		if (!myTokenizer.isDone()) {
			throw new RuntimeException("Trailing characters: "
					+ myTokenizer.getRemaining());
		}

		node.preprocessForGeneration();
		myScript = new ArrayList<IXissOperation>();
		node.generateCodeTo(myScript);

		return new XissQuery(myScript);
	}

	// Query = NonUnionQuery | QueryUnion
	private AstNode parseQuery(boolean mustSucceed) throws QueryParseException {
		AstNode res = null;
		if ((res = parseQueryUnion(false)) != null) {
			return res;
		}
		res = parseNonUnionQuery(mustSucceed);
		return res;
	}

	// NonUnionQuery = GlobalPath | (Query)
	private AstNode parseNonUnionQuery(boolean mustSucceed)
			throws QueryParseException {
		AstNode res = null;
		if (myTokenizer.eat('(')) {
			res = parseQuery(true);
			myTokenizer.eat(')', true);
			return res;
		} else if ((res = parseGlobalPath(false)) != null) {
			return res;
		}
		if (mustSucceed) {
			throw new QueryParseException("Query expected");
		}
		return null;
	}

	// GlobalPath = AbsPath | AnyLocPath
	// AbsPath = '/' [Path]
	// AnyLocPath = '//' [Path] | Path
	private GlobalPathNode parseGlobalPath(boolean mustSucceed)
			throws QueryParseException {
		myTokenizer.begin();

		boolean startWithRoot;
		boolean haveSlash = true;
		if (myTokenizer.eat("//", false)) {
			startWithRoot = false;
		} else if (myTokenizer.eat('/')) {
			startWithRoot = true;
		} else {
			startWithRoot = false;
			haveSlash = false;
		}

		PathNode path = parsePath(haveSlash ? false : mustSucceed);
		if (!haveSlash && mustSucceed && path == null) {
			throw new QueryParseException("Path expected");
		}
		myTokenizer.commit();
		return new GlobalPathNode(path, startWithRoot);
	}

	// QueryUnion = NonUnionQuery '|' Query
	private UnionNode parseQueryUnion(boolean mustSucceed)
			throws QueryParseException {
		myTokenizer.begin();
		AstNode left = parseNonUnionQuery(mustSucceed);
		if (left == null) {
			myTokenizer.rollback();
			return null;
		}
		if (!myTokenizer.eat('|')) {
			if (mustSucceed) {
				throw new QueryParseException("'|' expected");
			}
			myTokenizer.rollback();
			return null;
		}
		AstNode right = parseQuery(mustSucceed);
		if (right == null) {
			myTokenizer.rollback();
			return null;
		}
		myTokenizer.commit();
		return new UnionNode(left, right);
	}

	// Path = /* empty */ | PathChunk ['/' Path]
	private PathNode parsePath(boolean mustSucceed) throws QueryParseException {
		PathChunkNode chunk = parsePathChunk(false);
		if (chunk == null) {
			return null;
		}
		myTokenizer.begin();
		if (!myTokenizer.eat('/')) {
			myTokenizer.commit();
			return new PathNode(chunk);
		}

		PathNode right = parsePath(mustSucceed);
		if (right == null) {
			myTokenizer.rollback();
			return null;
		}
		myTokenizer.commit();
		return new PathNode(chunk, right);
	}

	// PathChunk = SingleChunk [Predicate] ['*']
	private PathChunkNode parsePathChunk(boolean mustSucceed)
			throws QueryParseException {
		// TODO

		PredicateNode predicate;
		AstNode core;
		boolean isKleeneClosure;

		core = parseChunkCore(mustSucceed);
		if (core == null) {
			return null;
		}
		predicate = parsePredicate(false);
		isKleeneClosure = myTokenizer.eat('*');
		return new PathChunkNode(core, predicate, isKleeneClosure);
	}

	/*
	 * SingleChunk = '_' | {Name} | '@'{Name} | '(' PathInBraces ')'
	 * PathInBraces = Path ['|' Path]
	 */
	private AstNode parseChunkCore(boolean mustSucceed)
			throws QueryParseException {
		myTokenizer.begin();
		if (myTokenizer.eat('_')) {
			myTokenizer.commit();
			return new AnyNode();
		}

		if (myTokenizer.eat('(')) {
			PathNode chunk = parsePath(mustSucceed);
			AstNode res = chunk;
			while (myTokenizer.eat('|')) {
				PathNode right = parsePath(mustSucceed);
				if (right == null) {
					myTokenizer.rollback();
					return null;
				}
				res = new UnionNode(res, right);
			}
			myTokenizer.eat(')', true);
			myTokenizer.commit();
			return res;
		}

		if (myTokenizer.eat('@', false)) {
			String name = myTokenizer.eatRegex(XissQueryParser.NAME_REGEX,
					mustSucceed);
			if (name == null) {
				myTokenizer.rollback();
				return null;
			}

			myTokenizer.commit();
			return new AttributeChunkNode(name);
		}

		String name = myTokenizer.eatRegex(XissQueryParser.NAME_REGEX,
				mustSucceed);
		if (name == null) {
			myTokenizer.rollback();
			return null;
		}

		myTokenizer.commit();
		return new ElementChunkNode(name);
	}

	/*
	 * Predicate = '[' (NodePredicate | AttributePredicate) ']' NodePredicate =
	 * {Name} AttributePredicate = '@' {Name} ['=' '"' {AttributeValue} '"']
	 */
	private PredicateNode parsePredicate(boolean mustSucceed)
			throws QueryParseException {
		if (!myTokenizer.eat('[')) {
			return null;
		}
		myTokenizer.begin();
		if (myTokenizer.eat('@')) {
			String name = myTokenizer.eatRegex(XissQueryParser.NAME_REGEX,
					mustSucceed);
			String value = null;
			if (myTokenizer.eat('=')) {
				if (!myTokenizer.eatAny("\"'", mustSucceed)) {
					myTokenizer.rollback();
					return null;
				}
				value = myTokenizer.eatRegex(
						XissQueryParser.ATTRIBUTE_VALUE_REGEX, mustSucceed);
				if (value == null) {
					myTokenizer.rollback();
					return null;
				}
				if (!myTokenizer.eatAny("\"'", mustSucceed)) {
					myTokenizer.rollback();
					return null;
				}
			}
			if (!myTokenizer.eat(']', mustSucceed)) {
				myTokenizer.rollback();
				return null;
			}
			myTokenizer.commit();
			return new AttributePredicateNode(name, value);
		} else {
			String name = myTokenizer.eatRegex(XissQueryParser.NAME_REGEX,
					mustSucceed);
			if (name == null) {
				myTokenizer.rollback();
				return null;
			}
			if (!myTokenizer.eat(']', mustSucceed)) {
				myTokenizer.rollback();
				return null;
			}
			myTokenizer.commit();
			return new ElementPredicateNode(name);
		}
	}
}
